from utils import *
import numpy as np
from ._distStruct import _DistStruct


#Compute the distance from each link to an object, for the current configuration
#of the robot
def _computehgDist(self, obj, h, g, q = None, htm = None, oldDistStruct=None, tol = 0.0005, noIterMax = 20):

  n = len(self.links)

  if q is None:
    q = self.q

  if htm is None:
    htm = self.MTH

  #Error handling
  if not Utils.isaVector(q,n):
    raise Exception("The optional parameter 'q' should be a "+str(n)+" dimensional vector")

  if not Utils.isaMatrix(htm,4,4):
    raise Exception("The optional parameter 'htm' should be a 4x4 homogeneous transformation matrix")
  
  if not(oldDistStruct is None):
    try:
      idRobot = oldDistStruct.idRobot
      idObj = oldDistStruct.idObj
      if not ( idObj == id(obj) and idRobot == id(self)):
        Exception("The optional parameter 'oldDistStruct' is a 'DistStruct' object, but it "\
        "must have to be relative to the SAME robot object and SAME external object, and "\
        "this is not the case")
    except:
      raise Exception("The optional parameter 'oldDistStruct' must be a 'DistStruct' object")
  
  #end error handling



  distStruct = _DistStruct(obj,self)

  J, FK = self.jacGeo(q, "dh", htm)

  colObjectCopy=[]

  #Update all collision objects of all links
  for i in range(n):
    colObjectCopy.append([])
    for j in range(len(self.links[i].colObject)):
      tempCopy = self.links[i].colObject[j][0].copy()
      htmd = self.links[i].colObject[j][1]
      tempCopy.setConfig(FK[i,:,:] @ htmd)
      colObjectCopy[i].append(tempCopy)
      #self.links[i].colObject[j][0].addConfig(FK[i,:,:] @ htmd)

  #Compute the distance structure
  for i in range(n):
    for j in range(len(self.links[i].colObject)):

      if oldDistStruct is None:
        pObj0 = np.random.uniform(-100,100, size=(3,))  
      else:
        pObj0 = oldDistStruct.getItem(i,j)["pObj"]


      pObj, pObjCol, d = Utils.computehgDist(obj, colObjectCopy[i][j], h, g, pObj0, tol, noIterMax)      
      
      JobjCol = J[i,0:3,:]-Utils.S(pObjCol-FK[i,0:3,3]) @ J[i,3:6,:]
      Jd = (np.transpose(pObjCol-pObj) @ JobjCol)/d

      distStruct._append(i, j, d, pObj, pObjCol, Jd)
      #st = {
      #  "linkNumber": i,
      #  "linkColObjNumber": j,
      #  "hgDistance": d,
      #  "pObj": pObj,
      #  "pObjCol": pObjCol,
      #  "jacDist": Jd
      #}

      #listStruct.append(st)
  
  return distStruct


import os, sys
currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)

import numpy as np
from utils import *
from links import *
from ._FK import _FK
from ._IK import _IK
from ._setConfig import _setConfig
from ._addConfig import _addConfig
from ._jacGeo import _jacGeo
from ._dynM import _dynM
from ._dynC import _dynC
from ._dynG import _dynG
from ._dynParam import _dynParam
from ._sDOF import _sDOF
from ._computehgDist import _computehgDist
from ._genCode import _genCode
from ._taskFunction import _taskFunction


class Robot:
	"""
  A class that contains a robot object in jupyterbot.

  Parameters
  ----------
  linkInfo : list of lists
      Contains the link info in a 6 x n list of lists, in which n is the
      number of joints.

      The first line is the "theta" parameter of Denavit-Hartenberg;

      The second line is the "d" parameter of Denavit-Hartenberg;

      The third line is the "alpha" parameter of Denavit-Hartenberg;

      The fourth line is the "a" parameter of Denavit-Hartenberg; 

      The fifth line is the joint type (0=revolute, 1 = prismatic);

      The sixth line is the configuration shift (sets the "zero" configuration
      of the robot).      

  htm : 4x4 numpy array or 4x4 nested list
      The robot base's configuration
      (default: 4x4 identity matrix).
  name : string
      The robot name 
      (default: 'genRobot').
  rtype : string
      The robot type
      (default: 'generic').
  colModel : list of list
      A list that contains at most n lists (n is the number of joints). Each
      list contains objects as balls, cylinders and boxs that compose the 
      collision model of the robot.
      (default: empty list).      
  """

  #######################################
  # Attributes
  #######################################


	@property
	def q(self):
		"""The current joint configuration."""
		return np.array(self._q)

	@property
	def htm(self):
		"""
		The current base configuration. 
		A 4x4 homogeneous matrix written is scenario coordinates.
		"""
		return np.array(self._htm)

	@property
	def links(self):
		"""Data structures containing the links of the robot."""
		return self._links

	@property
	def name(self):
		"""Name of the object."""
		return self._strName
  
  #######################################
  # Constructor
  #######################################

	_ROBOTTYPE = ["generic","sDOF"]
  
	def __init__(self, linkInfo, htm = np.identity(4), name = "genRobot", rtype = "generic", colModel = None ):
    #Error handling

		if not (str(type(linkInfo)) == "<class 'list'>"):
			raise Exception("The parameter 'linkInfo' should be a list with 6 with sublists, each with n float parameters")
		else:
			if not (len(linkInfo) == 6):
				raise Exception("The parameter 'linkInfo' should be a list 6 with sublists, , each with n float parameters")
			else:
				try:
					np.array(linkInfo, dtype=object)
				except:
					raise Exception("The parameter 'linkInfo' should be a list 6 with sublists, , each with n float parameters")


		if colModel is None:
			colModel = [[]]

		n = len(linkInfo[0])

		for i in range(n):
			if not(linkInfo[4][i]==0 or linkInfo[4][i]==1):
				raise Exception("The fifth list in 'linkInfo' should have either 0 (revolute) or 1 (prismatic)")


		if not (str(type(colModel)) == "<class 'list'>"):
			raise Exception("The optional parameter 'colModel' should be a list with at most "+str(n)+" sublist, each with ball, boxes or cylinders")
		else:
			if not (len(colModel) <= n):
				raise Exception("The optional parameter 'colModel' should be a list with at most "+str(n)+" sublist, each with ball, boxes or cylinders")
			else:
				for i in range(len(colModel)):
					try:
						for j in range(len(colModel[i])):
							if not Utils.isaSimpleObject(colModel[i][j]):
								raise Exception("The optional parameter 'colModel' should be a list with at most "+str(n)+" sublist, each with ball, boxes or cylinders")
					except:
						raise Exception("The optional parameter 'colModel' should be a list with at most "+str(n)+" sublist, each with ball, boxes or cylinders")            

		if not Utils.isaMatrix(htm,4,4):
			raise Exception("The optional parameter 'htm' should be a 4x4 homogeneous transformation matrix")

		if not (str(type(name)) == "<class 'str'>"):
			raise Exception("The optional parameter 'name' should be a string")

		if not (rtype in Robot._ROBOTTYPE):
			raise Exception("The optional parameter 'rtype' type can only be one of the following" \
      "strings: "+str(Robot._ROBOTTYPE))  

    #end error handling

		self._frames = []
		self._linkInfo = linkInfo
		self._strlinkInfo = str(self._linkInfo)    
		self._htm = np.array(htm)
		self._strName = name
		self._rtype = rtype


		self._links = []
		for i in range(n):
			self._links.append(
        Link(self,i,
             self._linkInfo[0][i],
             self._linkInfo[1][i],
             self._linkInfo[2][i],
             self._linkInfo[3][i],
             self._linkInfo[4][i],
             self._linkInfo[5][i]))
			for j in range(len(colModel[i])):
				self._links[i].attachColObject(colModel[i][j],colModel[i][j].htm)	

    #Recover initial configuration and shift configuration from DH table    

		self._q0 = np.zeros((n,))
		self._dq0 = np.reshape(self._linkInfo[5],(n,))

		for i in range(n):
			if  self._links[i].jointType==0:
				self._q0[i] = self._links[i].theta-self._dq0[i]

			if self._links[i].jointType==1:
				self._q0[i] = self._links[i].d-self._dq0[i]


    #Set initial total configuration
		self.setConfig(self._q0,self._htm)

  #######################################
  # Std. Print
  #######################################

	def __repr__(self):
		n = len(self._linkInfo[0])

		string = "Robot with name '"+self._strName+"': \n\n"
		string += " Number of joints: "+str(n)+"\n"
		string += " Joint types: "

		for i in range(n):
			string += "R" if self._links[i].jointType==0 else "P"
    
		string+="\n"
		string+=" Current configuration: "+str( [round(num, 3) for num in np.ndarray.tolist(self._q)] )+"\n"
		string+=" Current base HTM: \n"+str(self._htm)+"\n"
		string+=" Current end-effector HTM: \n"+str(self.FK())
		return string

  #######################################
  # Methods for configuration changing
  #######################################


	def addConfig(self, q, htm = None):
		"""
    Add a single configuration to the object's animation queue.

    Parameters
    ----------
    q : nd numpy vector or array
        The manipulator's joint configuration.
    htm : 4x4 numpy array or 4x4 nested list
        The robot base's configuration
        (default: the same as the current HTM).

    Returns
    -------
    None
    """
		return _addConfig(self, q, htm)

	def setConfig(self, q = None, htm = None):
		"""
    Reset object's animation queue and add a single configuration to the 
    object's animation queue.

    Parameters
    ----------
    q : nd numpy vector or array
        The manipulator's joint configuration 
        (default: the default joint configuration for the manipulator).
    htm : 4x4 numpy array or 4x4 nested list
        The robot base's configuration
        (default: the same as the current HTM).

    Returns
    -------
    None
    """    
		return _setConfig(self, q, htm)

  #######################################
  # Methods for kinematics model
  #######################################

	def FK(self, q = None, axis = 'eef', htm = None):
		"""
    Compute the forward kinematics for an axis at a given joint and base
    configuration. Everything is written in the scenario coordinates.

    Parameters
    ----------
    q : nd numpy vector or array
        The manipulator's joint configuration 
        (default: the default joint configuration for the manipulator).
    axis : string
        For which axis you want to compute the FK:
        'eef': for the end-effector
        'dh': for all Denavit-Hartenberg axis
        'com': for all center-of-mass axis
        (default: 'eef').    
    htm0 : 4x4 numpy array or 4x4 nested list
        The robot base's configuration
        (default: the same as the current HTM).

    Returns
    -------
    FK : 4x4 or nx4x4 numpy matrix
        For axis='eef', returns a single htm. For the other cases, return
        n htms as a nx4x4 numpy matrix.
    """    
		return _FK(self, q, axis, htm0)

	def IK(self, htm, q0 = None, ptol=0.005, atol=5, noIterMax=2000):
		"""
    Try to solve the inverse kinematic problem for the end-effector, given a
    desired homogeneous transformation matrix. It returns the manipulator
    configuration. 

    Uses an iterative algorithm.

    The algorithm can fail, throwing an Exception when it happens.

    Parameters
    ----------
    htm0 : 4x4 numpy array or 4x4 nested list
        The desired end-effector HTM, written in scenario coordinates    
    q0 : nd numpy vector or array
        Initial guess for the algorithm for the joint configuration
        (default: a random joint configuration).
    ptol : positive float
        The accepted error for the end-effector position, in meters
        (default: 0.005 m).    
    atol : positive float
        The accepted error for the end-effector orientation, in degrees
        (default: 5 degrees). 
    noIterMax : positive int
        The maximum number of iterations for the algoritm
        (default: 2000 iterations). 

    Returns
    -------
    q : nd numpy vector or array
        The configuration that solves the IK problem
    """    
		return _IK(self, htm, q0, ptol, atol, noIterMax)

	def jacGeo(self, q = None, axis = 'eef', htm0 = None):
		"""
    Compute the geometric Jacobian for an axis at a given joint and base
    configuration. Also returns the forward kinematics as a by-product.
    Everything is written in the scenario coordinates.

    Parameters
    ----------
    q : nd numpy vector or array
        The manipulator's joint configuration 
        (default: the default joint configuration for the manipulator).
    axis : string
        For which axis you want to compute the FK:
        'eef': for the end-effector
        'dh': for all Denavit-Hartenberg axis
        'com': for all center-of-mass axis
        (default: 'eef').    
    htm0 : 4x4 numpy array or 4x4 nested list
        The robot base's configuration 
        (default: the same as the current htm).

    Returns
    -------
    J : 6xn or nx6xn numpy matrix
        For axis='eef', returns a single 6xn Jacobian. For the other cases, 
        return n Jacobians as a nx6xn numpy matrix.

    FK : 4x4 or nx4x4 numpy matrix
        For axis='eef', returns a single htm. For the other cases, return
        n htms as a nx4x4 numpy matrix.
    """ 
		return _jacGeo(self, q, axis, htm0)


  #######################################
  # Methods for dynamics model
  #######################################

	def dynM(self, q = None):
		"""
    Compute the generalized inertia matrix at a given joint configuration.

    Parameters
    ----------
    q : nd numpy vector or array
        The manipulator's joint configuration, a nD numpy vector or nD array 
        (default: the default joint configuration for the manipulator).
    Returns
    -------
    M : nxn numpy array
        The generalized inertia matrix at the joint configuration q.
    """    
		return _dynM(self, q)

	def dynC(self, qdot, q = None):
		"""
    Compute the generalized Coriolis-Centrifugal torques at a given joint 
    configuration speed and joint configuration.

    Parameters
    ----------
    qdot : nd numpy vector or array
        The manipulator's joint configuration speed. 

    q : nd numpy vector or array
        The manipulator's joint configuration 
        (default: the default joint configuration for the manipulator).

    Returns
    -------
    C : nD numpy vector
        The generalized Coriolis-Centrifugal torques at the joint 
        configuration q and joint configuration speed qdot.
    """     
		return _dynC(self, qdot, q)

	def dynG(self, q = None):
		"""
    Compute the generalized gravity torques at a given joint configuration.

    Parameters
    ----------
    q : nd numpy vector or array
        The manipulator's joint configuration, a nD numpy vector or nD array 
        (default: the default joint configuration for the manipulator).
    Returns
    -------
    G : nD numpy vector
        The generalized gravity torques at the joint configuration q.
    """        
		return _dynG(self, q)

	def dynParam(self, qdot, q = None):
		"""
    Compute all the three dynamic parameters at a given joint configuration
    and joint configuration speed. It is more efficient than calling them
    separatedly.

    Parameters
    ----------
    qdot : nd numpy vector or array
        The manipulator's joint configuration speed .

    q : nd numpy vector or array
        The manipulator's joint configuration 
        (default: the default joint configuration for the manipulator).

    Returns
    -------
    M : nxn numpy array
        The generalized inertia matrix at the joint configuration q.

    G : nD numpy vector
        The generalized gravity torques at the joint configuration q.

    C : nD numpy vector
        The generalized Coriolis-Centrifugal torques at the joint 
        configuration q and joint configuration speed qdot.
    """    
		return _dynParam(self, qdot, q)


  #######################################
  # Methods for control
  #######################################

	def taskFunction(self, FKd, q = None, htm = None):
		"""
    6-dimensional task function for end-effector pose control, given a joint 
    configuration, a base configuration and the desired pose FKd. Everything 
    is written in scenario coordinates.
    Also returns the Jacobian of this function.

    Parameters
    ----------
    Fkd : 4x4 numpy array or 4x4 nested list
        The desired end-effector pose. 
 
    q : nd numpy vector or array
        The manipulator's joint configuration 
        (default: the default joint configuration for the manipulator).

    htm : 4x4 numpy array or 4x4 nested list
        The robot base's configuration 
        (default: the same as the current htm).

    Returns
    -------
    r : 6-dimensional numpy vector
        The task function.

    Jr : 6xn numpy matrix
        The respective task Jacobian.
    """    
		return _taskFunction(self, FKd, q, htm)

  #######################################
  # Methods for simulation
  #######################################

	def genCode(self):
		"""Generate code for injection."""   
		_genCode(self)

  #######################################
  # Robot constructors
  #######################################

	def sDOF(htm = np.identity(4), name = 'sdofRobot'):
		"""
    Create a six-degree of freedom manipulator.

    Parameters
    ----------
    htm : 4x4 numpy array or 4x4 nested list
        The initial base configuration for the robot. 
 
    name : string
        The robot name
        (default: 'sdofRobot').

    htm : 4x4 numpy array or 4x4 nested list
        The robot base's configuration 
        (default: the same as the current htm).

    Returns
    -------
    R : Robot object
        The robot.

    """    
		linkInfo, colModel = _sDOF(htm, name)
		return Robot(linkInfo, htm, name, 'sDOF',colModel)

  #######################################
  # Advanced methods
  #######################################

	def computehgDist(self, obj, h=0.1, g=0.1, q = None, htm = None, oldDistStruct=None, tol = 0.0005, noIterMax = 20):
		"""
    Compute (h,g) distance structure from each one of the robot's link to a 
    'simple' external object (ball, box or cylinder), given a joint and base 
    configuration.

    Use an iterative algorithm, based no h-projections and a modification of 
    Von Neumann's cyclic projection algorithm.

    Parameters
    ----------
    obj : a simple object (ball, box or cylinder)
        The external object for which the distance structure is going to be 
        computed, for each robot link.
    h : positive float
        Smoothing parameter for the robot's links, in meters
        (default: 0.1 m).    
    g : positive float
        Smoothing parameter for the external object
        (default: 0.1 m).
    q : nd numpy vector or array
        The manipulator's joint configuration 
        (default: the default joint configuration for the manipulator).
    htm : 4x4 numpy array or 4x4 nested list
        The robot base's configuration 
        (default: the same as the current htm).
    oldDistStruct : 'DistStruct' object
        'DistStruct' obtained previously for the same robot and external object.
        Can be used to enhance the algorith speed using the previous closest 
        point as an initial guess
        (default: None).
    tol : positive float
        Tolerace for convergence in the iterative algorithm, in meters.
        (default: 0.0005 m).        
    noIterMax : positive int
        The maximum number of iterations for the algoritm
        (default: 20 iterations). 

    Returns
    -------
    distStruct : 'DistStruct' object
        Distance struct for each one of the m objects that compose the robot's
        collision model. Contais m dictionaries. Each one of these dictionaries 
        contains the following entries:
        
        'linkNumber', containing the robot link index for this entries

        'linkColObjNumber', containing the link index for the covering object 
          at that link

        'hgDistance', containing the smoothed hg distance

        'pObj', containing the closest point at the external object

        'pObjCol', containing the closest point at the covering object in the 
          link
        
        'jacDist', containing the Jacobian (gradient) of this distance function       
    """

		return _computehgDist(self, obj, h, g, q, htm, oldDistStruct, tol, noIterMax)



import numpy as np
from utils import *


class Link:

  #######################################
  # Attributes
  #######################################
  
	@property
	def theta(self):
    """The 'theta' parameter of the Denavit-Hartenberg convention (in rad)"""
		return self._theta

	@property
	def d(self):
    """The 'd' parameter of the Denavit-Hartenberg convention (in meters)"""    
		return self._d

	@property
	def a(self):
    """The 'a' parameter of the Denavit-Hartenberg convention (in meters)"""        
		return self._a

	@property
	def alpha(self):
    """The 'alpha' parameter of the Denavit-Hartenberg convention (in rad)"""     
		return self._alpha

	@property
	def jointNumber(self):
    """The joint number in the kinematic chain."""      
		return self._jointNumber

	@property
	def jointType(self):
    """The joint type (0=revolute, 1=prismatic)."""        
		return self._jointType

	@property
	def centerShift(self):
    """???"""     
		return self._centerShift

	@property
	def mass(self):
    """The link's mass, in kg."""        
		return self._mass

	@property
	def inertiaMatrix(self):
    """The link's inertia matrix, in kg m²."""           
		return self._inertiaMatrix

	@property
	def modelPath(self):
    """Path to the geometric model."""       
		return self._modelPath

	@property
	def colObject(self):
    """Collection of objects that compose the collision model of this link.""" 
		return self._colObject

  #######################################
  # Constructor
  #######################################
      
	def __init__(self, robot, jointNumber, theta, d, alpha, a, jointType,
  jointShift=0, centerShift = np.zeros((3,)), mass=1, inertiaMatrix = np.identity(3), modelPath=""  ):

    #Error handling
		if str(type(robot)) != "<class 'jupyterbot.robot.robot.Robot'>":
			raise Exception("The 'robot' parameter should be a robot handle")

		if str(type(jointNumber)) != "<class 'int'>" or jointNumber<0:
			raise Exception("The 'jointNumber' parameter should be a nonnegative integer")

		if not Utils.isaNumber(theta):
			raise Exception("The 'theta' parameter should be a float")

		if not Utils.isaNumber(d):
			raise Exception("The 'd' parameter should be a float")

		if not Utils.isaNumber(alpha):
			raise Exception("The 'alpha' parameter should be a float")

		if not Utils.isaNumber(a):
			raise Exception("The 'a' parameter should be a float")

		if jointType != 0 and jointType != 1:
			raise Exception("The 'jointType' parameter should be either '0' (rotative) or '1' (prismatic)")

    #Code

		self._robot = robot
		self._jointNumber = jointNumber
		self._theta = theta
		self._d = d
		self._alpha = alpha
		self._a = a
		self._jointType = jointType
		self._colObject = []

		self._jointShift = jointShift
		self._centerShift = np.array(centerShift)
		self._mass = mass 
		self._inertiaMatrix = np.array(inertiaMatrix)
		self._modelPath = modelPath


    #Error handling
		if not Utils.isaNumber(self._jointShift):
			raise Exception("The parameter 'jointShift' should be a float")

		if not Utils.isaVector(self._centerShift,3):
			raise Exception("The parameter 'centerShift' should be a 3D array") 

		if not Utils.isaNumber(self._mass) or self._mass<0 :
			raise Exception("The parameter 'mass' should be a positive float")
          
		if not Utils.isaPDMatrix(self._inertiaMatrix,3):
			raise Exception("The parameter 'inertiaMatrix' should be a symmetric positive definite 3x3 matrix")

		if not (str(type(self._modelPath)) == "<class 'str'>"):
			raise Exception("The parameter 'modelPath' should be a string")

  #######################################
  # Std. Print
  #######################################

	def __repr__(self):
    
		string = "Link "+str(self._jointNumber)+" of robot '"+str(self._robot.name)+"' "

		if self._jointType == 0:
			string += "with rotative joint:\n\n"
			string += " θ (rad): "+str(round(self._robot.q[self._jointNumber],3))+" [variable] \n"
			string += " d (m)  : "+str(self._d)+" \n"
		if self._jointType == 1:
			string += "with prismatric joint:\n\n" 
			string += " θ (rad): "+str(self._theta)+" \n"
			string += " d (m)  : "+str(round(self._robot.q[self._jointNumber],3))+" [variable] \n"
    
		string += " α (rad): "+str(self._alpha)+" \n"
		string += " a (m)  : "+str(self._a)+" \n"
		string += " Link mass (kg): "+str(self._mass)+"\n"  
		string += " Link inertia matrix (kg*m²): \n "+str(self._inertiaMatrix)+"\n" 
		string += " Link model path: "+str(self._modelPath)+"\n"

		return string

  #######################################
  # Methods
  #######################################
  
	def attachColObject(self, obj, htm):

    #Error handling
		if not Utils.isaMatrix(htm,4,4):
			raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix")


		if not( Utils.isaSimpleObject(obj) ):
			raise Exception("The parameter 'obj' must be either a box, a ball or a cylinder")

		self._colObject.append([obj,htm])


from utils import *
import numpy as np

#Geometric Jacobian
def _jacGeo(self, q = None, axis = 'eef', htm0 = None):

  if q is None:
    q = self.q
  if htm0 is None:
    htm0 = self.htm

  n = len(self._linkInfo[0])

  #Error handling
  if not Utils.isaVector(q,n):
    raise Exception("The optional parameter 'q' should be a "+str(n)+" dimensional vector")

  if not (axis=="eef" or axis=="dh" or axis=="com"):
    raise Exception("The optional parameter 'axis' should be one of the following strings:\n"\
    "'eef': End-effector \n"\
    "'dh': All "+str(n)+" axis of Denavit-Hartenberg\n"\
    "'com': All "+str(n)+" axis centered at the center of mass of the objects")

  if not Utils.isaMatrix(htm0,4,4):
    raise Exception("The optional parameter 'htm' should be a 4x4 homogeneous transformation matrix")
  #end error handling

  if axis=='dh' or axis=='eef':
    T = self.FK(q,'dh',htm0)
  if axis=='com':
    T = self.FK(q,'com',htm0)

  J = np.zeros((n,6,n))
  

  for i in range(n):
    pi = T[i,0:3,3]
    for j in range(i+1):
    
      if j == 0:
        pjant = htm0[0:3,3]
        zjant = htm0[0:3,2]
      else:
        pjant = T[j-1,0:3,3]
        zjant = T[j-1,0:3,2]

      if self._links[i].jointType==0:
        J[i,0:3,j] = Utils.S(zjant) @ (pi - pjant)
        J[i,3:6,j] = zjant 
    
      if self._links[i].jointType==1:
        J[i,0:3,j] = zjant 
        J[i,3:6,j] = [[0],[0],[0]]

  #GAMBIARRA
  for i in range(n):
    J[i,3:6,5]=-J[i,3:6,5]

  if axis=='dh' or axis=='com':
    return J,T
  
  if axis=='eef':
    return J[-1,:,:], T[-1,:,:]



from utils import *
import numpy as np

class Cylinder:
  """
  A cylinder object.

  Parameters
  ----------
  htm : 4x4 numpy array or 4x4 nested list
      The object's configuration
      (default: the same as the current HTM).

  name : string
      The object's name
      (default: 'genBall').

  radius : positive float
      The cylinder base radius, in meters.
      (default: 1).    

  height : positive float
      The cylinder's height, in meters
      (default: 1).  

  mass : positive float
      The object's mass, in kg
      (default: 1).  

  color : string
      The object's color, in HTML compatible string
      (default: "red").              
  """

  #######################################
  # Attributes
  #######################################

  @property
  def radius(self):
    """The cylinder base radius, in meters."""
    return self._radius

  @property
  def height(self):
    """The cylinder height, in meters."""
    return self._height

  @property
  def name(self):
    """The object name."""     
    return self._name

  @property
  def htm(self):
    """
    Object pose. 
    A 4x4 homogeneous transformation matrix written is scenario coordinates.
    """     
    return np.array(self._htm)  

  @property
  def mass(self):
    """Mass of the object, in kg.""" 
    return self._mass 

  @property
  def color(self):
    """Color of the object, HTML compatible""" 
    return self._color
  #######################################
  # Constructor
  #######################################
  
  def __init__(self, htm = np.identity(4), name="genCyl", radius=1, height=1, mass=1, color="red"):

    #Error handling
    if not Utils.isaMatrix(htm,4,4):
      raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix")

    if not Utils.isaNumber(mass) or mass<0 :
      raise Exception("The parameter 'mass' should be a positive float")

    if not Utils.isaNumber(radius) or radius<0 :
      raise Exception("The parameter 'radius' should be a positive float")

    if not Utils.isaNumber(height) or height<0 :
      raise Exception("The parameter 'height' should be a positive float")

    if not (str(type(name)) == "<class 'str'>"):
      raise Exception("The parameter 'name' should be a string")     

    if not Utils.isaColor(color):
      raise Exception("The parameter 'color' should be a color") 
    #end error handling
    
    self._radius = radius
    self._height = height
    self._color = color
    self._htm = np.array(htm)
    self._strColor = "'" + str(self._color) + "'"
    self._name = name
    self._strName = "'" + name + "'"
    self._mass = 1
    self._isaSimpleObject = True
    
    #Set initial total configuration
    self.setConfig(self._htm)

  #######################################
  # Std. Print
  #######################################

  def __repr__(self):

    string = "Cylinder with name '"+self._name+"': \n\n"
    string += " Radius (m): "+str(self._radius)+"\n"
    string += " Height (m): "+str(self._height)+"\n"
    string += " Color: "+str(self._color)+"\n"    
    string += " Mass (kg): "+str(self._mass)+"\n"
    string += " HTM: \n"+str(self._htm)+"\n"
    
    return string


  #######################################
  # Methods
  #######################################


  def addConfig(self, htm = None):
    """
    Add a single configuration to the object's animation queue.

    Parameters
    ----------
    htm : 4x4 numpy array or 4x4 nested list
        The object's configuration
        (default: the same as the current HTM).

    Returns
    -------
    None
    """       
    if htm is None:
      htm = self._htm

    #Error handling
    if not Utils.isaMatrix(htm,4,4):
      raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix")
    #end error handling

    f = [ htm[0][0], htm[0][2], -htm[0][1], htm[0][3],
          htm[1][0], htm[1][2], -htm[1][1], htm[1][3],
          htm[2][0], htm[2][2], -htm[2][1], htm[2][3],
          0        ,         0,         0,         1]

    self._htm = htm
    self._frames.append(f)

  def setConfig(self, htm = None):
    """
    Reset object's animation queue and add a single configuration to the 
    object's animation queue.

    Parameters
    ----------
    htm : 4x4 numpy array or 4x4 nested list
        The object's configuration
        (default: the same as the current HTM).

    Returns
    -------
    None
    """
    if htm is None:
      htm = self._htm

    #Error handling
    if not Utils.isaMatrix(htm,4,4):
      raise Exception("The optional parameter 'htm' should be a 4x4 homogeneous transformation matrix")
    #end error handling      

    self._frames = []
    self.code = ''
    self.addConfig(htm)

  def genCode(self):
    """Generate code for injection.""" 
    self._strFrames = str(self._frames)
    self.code ='''
		const ''' + self._name + ''' = new Cylinder(''' + self._strName + ''',''' + str(self._radius) + ''',''' + str(self._height) + ''',''' + self._strColor + ''',''' + self._strFrames + ''');
		sceneElements.push(''' + self._name + ''');
		//USER INPUT GOES HERE
		'''

  #Compute inertia matrix with respect to the inertia frame
  def inertiaMatrix(self, htm = None):
    """
    The 3D inertia matrix of the object, written in the 'inertia frame', that is, a frame attached to the center of mass of the object which has the same orientation as the scenario frame.

    Parameters
    ----------
    htm : 4x4 numpy array or 4x4 nested list
        The object's configuration for which the inertia matrix will be computed
        (default: the same as the current HTM).

    Returns
    -------
     inertiaMatrix : 3x2 numpy array
        The 3D inertia matrix.
    """  

    if htm is None:
      htm = self._htm

    #Error handling
    if not Utils.isaMatrix(htm,4,4):
      raise Exception("The optional parameter 'htm' should be a 4x4 homogeneous transformation matrix")
    #end error handling      

    Ixx = (1/12)*self._mass*(3*self._radius*self._radius + self._height*self._height)
    Iyy = (1/12)*self._mass*(3*self._radius*self._radius + self._height*self._height)   
    Izz = (1/2)*self._mass*(self._radius*self._radius)
    Q = htm[0:3,0:3]

    return np.transpose(Q) @ np.diag([Ixx, Iyy, Izz]) @ Q 

  def copy(self):
    """Return a deep copy of the object, without copying the animation frames."""    
    return Cylinder(self.htm, self.name+"_copy", self.radius, self.height, self.mass, self.color)
    
  #Compute the h projection of a point into an object
  def hProjection(self, point, h = 0.001, htm = None):
    """
    The h projection of a point in the object, that is, the 
    h-closest point in the object to a point 'point'.

    Parameters
    ----------
    point : 3D vector
        The point for which the projection will be computed.

    h : positive float
        Smoothing parameter, in meters
        (defalt: 0.001 m)

    htm : 4x4 numpy array or 4x4 nested list
        The object's configuration
        (default: the same as the current HTM).            

    Returns
    -------
     projPoint : 3D vector
        The h-projection of the point 'point' in the object.

     d : positive float
        The h-distance between the object and 'point'.     
    """  

    if htm is None:
      htm = self._htm

     #Error handling
    if not Utils.isaMatrix(htm,4,4):
      raise Exception("The optional parameter 'htm' should be a 4x4 homogeneous transformation matrix")

    if not Utils.isaVector(point,3):
      raise Exception("The parameter 'point' should be a 3D vector")

    if not Utils.isaNumber(h) or h<=0:
      raise Exception("The optional parameter 'h' should be a positive number")
    #end error handling        
    tpoint = np.transpose(htm[0:3,0:3]) @ (point - htm[0:3,3])

    delta = 0.001

    r = sqrt( tpoint[0]*tpoint[0] + tpoint[1]*tpoint[1])

    drf = Utils.funCir(r+delta,h, self.radius)
    drb = Utils.funCir(r-delta,h, self.radius)

    dzf = Utils.funInt(tpoint[2]+delta,h, 0.5*self.height)
    dzb = Utils.funInt(tpoint[2]-delta,h, 0.5*self.height)

    dr = (drf - drb)/(2*delta)

    x = tpoint[0] - dr*tpoint[0]/(r+0.00001)
    y = tpoint[1] - dr*tpoint[1]/(r+0.00001)    
    z = tpoint[2] - (dzf-dzb)/(2*delta)

    d = 0.5*(drf+drb+dzf+dzb) 

    return htm[0:3,0:3] @ np.array([x,y,z]) + htm[0:3,3], d      


from utils import *
import numpy as np

#Dynamic parameter Gravity
def _dynG(self, q = None):

  if q is None:
    q = self.q

  #Error handling
  if not Utils.isaVector(qdot,n):
    raise Exception("The parameter 'qdot' should be a "+str(n)+" dimensional vector")
  #end error handling  

  Jc, FKc = self.jacGeo(q,'com') 
  return self._dynGaux(qdot, q, Jc, FKc)    

def _dynGaux(self, q, Jc, FKc):

  n = len(self._links)
  G = np.zeros((n,))
  g = 9.8;
  for i in range(n):
    G = G + g * self._links[i].mass * np.transpose(Jc[i,3,:]),
  
  return np.reshape(G,(n,))


from utils import *
import numpy as np

#Dynamic parameter Coriolis-Centrifugal
def _dynC(self, qdot, q = None):

  if q is None:
    q = self.q

  #Error handling
  if not Utils.isaVector(qdot,n):
    raise Exception("The parameter 'qdot' should be a "+str(n)+" dimensional vector")

  if not Utils.isaVector(q,n):
    raise Exception("The optional parameter 'q' should be a "+str(n)+" dimensional vector")
  #end error handling  

  Jc, FKc = self.jacGeo(q,'com') 
  return self._dynCaux(qdot, q, Jc, FKc)

#Dynamic parameter Coriolis-Centrifugal
def _dynCaux(self, qdot, q, Jc, FKc):


  if (Jc is None) or (FKc is None):
    Jc, FKc = self.jacGeo(q,'com')

  #Error handling
  if not Utils.isaVector(qdot,n):
    raise Exception("The parameter 'qdot' should be a "+str(n)+" dimensional vector")

  if not Utils.isaVector(q,n):
    raise Exception("The optional parameter 'q' should be a "+str(n)+" dimensional vector")
  #end error handling


  M = self.dynM(q,Jc,FKc)
  n = len(self._links)

  dq=0.001
  dMdq = []
  for i in range(n):
    dqi = np.zeros((n,))
    dqi[i] = dq
    Mpi = self.dynM(q+dq)
    dMdq.append((Mpi-M)/dq)

  C1 = np.zeros((n,))
  C2 = np.zeros((n,))

  for j in range(n):
    C1 = C1 + (qdot[j]*dMdq[j]) @ qdot
    
  for i in range(n):
    C2[i] = 0.5 * np.transpose(qdot) @ dMdq[i] @ qdot

  return np.reshape(C1-C2,(n,))




from IPython.core.display import display, HTML
import re
import numpy as np
from .utils import *
from .robot import *
from .links import *

class Simulation:

  _STRJAVASCRIPT = '''
<canvas id="scene" width="800" height="600"></canvas>

<script type="module">
	import { Object3D, Vector3, BoxBufferGeometry, Color, Mesh, MeshBasicMaterial, PerspectiveCamera,
	  Scene, WebGLRenderer, AmbientLight, DirectionalLight, HemisphereLight, MeshStandardMaterial,
	  AxesHelper, GridHelper, Matrix4, SphereBufferGeometry, CylinderBufferGeometry, Group,
	} from "https://cdn.skypack.dev/three@0.132.2";
	import { OrbitControls } from 'https://threejs.org/examples/jsm/controls/OrbitControls.js';
	import {OBJLoader} from 'https://threejs.org/examples/jsm/loaders/OBJLoader.js';
  	
	//--------------------SIMULATION ELEMENTS---------------------
	class Objsim{
		constructor(_frames){
			this.frames = _frames;
			this.currentFrame = 0;
			this.shape = "I HAVE NO SHAPE YET";
		}
		
		nextFrame(){
			if(this.currentFrame < this.frames.length){
				this.shape.matrix.set( this.frames[this.currentFrame][ 0],this.frames[this.currentFrame][ 1],this.frames[this.currentFrame][ 2],this.frames[this.currentFrame][ 3],
								       this.frames[this.currentFrame][ 4],this.frames[this.currentFrame][ 5],this.frames[this.currentFrame][ 6],this.frames[this.currentFrame][ 7],
			                           this.frames[this.currentFrame][ 8],this.frames[this.currentFrame][ 9],this.frames[this.currentFrame][10],this.frames[this.currentFrame][11],
									   this.frames[this.currentFrame][12],this.frames[this.currentFrame][13],this.frames[this.currentFrame][14],this.frames[this.currentFrame][15]);
				this.currentFrame = this.currentFrame + 1;
			}
		}
	}
	
	class Box extends Objsim{
		constructor(_name, _width, _height, _depth, _color, _frames){
			super(_frames);
			this.width = _width;
			this.height = _height;
			this.depth = _depth;
			this.color = _color;
			const geometry = new BoxBufferGeometry( this.width, this.height, this.depth);
			const material = new MeshStandardMaterial( {color: this.color, transparent: true, opacity: 0.6} );
			const cube = new Mesh( geometry, material );
			cube.name = _name;
			cube.matrixAutoUpdate = false;
			this.shape = cube;
		}
	}
	
	class Ball extends Objsim{
		constructor(_name, _radius, _color, _frames){
			super(_frames);
			this.radius = _radius;
			this.color = _color;
			const geometry = new SphereBufferGeometry( this.radius, 32, 16);
			const material = new MeshStandardMaterial( { color: this.color, transparent: true, opacity: 0.6} );
			const sphere = new Mesh( geometry, material );
			sphere.name = _name;
			sphere.matrixAutoUpdate = false;
			this.shape = sphere;
		}
	}
	
	class Cylinder extends Objsim{
		constructor(_name, _radius, _height, _color, _frames){
			super(_frames);
			this.radius = _radius;
			this.height = _height;
			this.color = _color;
			const geometry = new CylinderBufferGeometry( this.radius, this.radius, this.height, 10 );
			const material = new MeshStandardMaterial( {color: this.color, transparent: true, opacity: 0.6} );
			const cylinder = new Mesh( geometry, material );
			cylinder.name = _name;
			cylinder.matrixAutoUpdate = false;
            this.shape = cylinder;
		}
	}
	
	class Robot extends Objsim{
		constructor(_linkInfo, _frames){
			super(_frames);
			//Function that creates a generic robot
			const base = new Group();
			base.name = "base";
			function createLinks(linkInfo, size){
				let links = [];
					for(let i = 0; i < linkInfo[0].length + 1; i++){
						const link = new Group();
						//link.matrixAutoUpdate = false;
						const axesHelper = new AxesHelper( size);
						axesHelper.matrixAutoUpdate = false;
						link.add(axesHelper)
						link.name = "link" + (i).toString();
						if(i != 0){
							link.rotateZ(linkInfo[0][i - 1]);
							link.translateZ(linkInfo[1][i - 1]);
							link.rotateX(linkInfo[2][i - 1]);
							link.translateX(linkInfo[3][i - 1]);
						}
						if(linkInfo[4][i] == 0){
							const geometry = new CylinderBufferGeometry( size/9, size/9, size/3, 20 );
							const material = new MeshStandardMaterial( {color: "blue"} );
							const cylinder = new Mesh( geometry, material );
							cylinder.rotateX(1.57);
							link.add(cylinder)
						}else if(linkInfo[4][i] == 1){
							const geometry = new BoxBufferGeometry( size/3.1, size/3.1, size/3.1);
							const material = new MeshStandardMaterial( {color: "red"} );
							const cube = new Mesh( geometry, material );
							link.add(cube)
						}else if(linkInfo[4][i] == 2){
							const geometry = new SphereBufferGeometry( size/5.5, 32, 16);
							const material = new MeshStandardMaterial( { color: "black" } );
							const sphere = new Mesh( geometry, material );
							link.add(sphere)
						}
						link.updateMatrix();
						links.push(link);
					}
					for(let i = 0; i < linkInfo[0].length; i++){
						links[i].add(links[i+1])
					}
				
					return links[0]
			};
			this.linkInfo = _linkInfo;
			this.shape = base;
			base.add(createLinks(this.linkInfo, 0.2));
			console.log(this.shape);
			this.shape.matrixAutoUpdate = false;
		}
		//Function that updates frames
		nextFrame(){
		
			if(this.currentFrame < this.frames.length){
				//setting robot position
				this.shape.matrix.set( this.frames[this.currentFrame][ 0],this.frames[this.currentFrame][ 1],this.frames[this.currentFrame][ 2],this.frames[this.currentFrame][ 3],
								       this.frames[this.currentFrame][ 4],this.frames[this.currentFrame][ 5],this.frames[this.currentFrame][ 6],this.frames[this.currentFrame][ 7],
			                           this.frames[this.currentFrame][ 8],this.frames[this.currentFrame][ 9],this.frames[this.currentFrame][10],this.frames[this.currentFrame][11],
									   this.frames[this.currentFrame][12],this.frames[this.currentFrame][13],this.frames[this.currentFrame][14],this.frames[this.currentFrame][15]);
				//setting robot configuration
				if(this.frames[this.currentFrame][16] != undefined){
					let linkName = "";
					let j = 0;
					for(let i = 0; i < this.linkInfo[4].length; i++){
						linkName = "link" + (i).toString();
						if(this.linkInfo[4][i] == 0){
							this.shape.getObjectByName(linkName).rotation.z = this.frames[this.currentFrame][16][j];
							j++;
						}else if (this.linkInfo[4][i] == 1){
							this.shape.getObjectByName(linkName).position.z = this.frames[this.currentFrame][16][j];
							j++;						
						}
					}
				}
				//grabing stuff
				if(this.frames[this.currentFrame][17] != undefined){
					let linkName = "link" + (this.linkInfo[4].length - 1).toString();
					console.log(linkName);
					let newFrameX = this.shape.getObjectByName(linkName).matrixWorld.elements[12];
					let newFrameY = this.shape.getObjectByName(linkName).matrixWorld.elements[13];
					let newFrameZ = this.shape.getObjectByName(linkName).matrixWorld.elements[14];
					let oldFrameX = scene.getObjectByName(this.frames[this.currentFrame][17][1]).matrixWorld.elements[12];
					let oldFrameY = scene.getObjectByName(this.frames[this.currentFrame][17][1]).matrixWorld.elements[13];
					let oldFrameZ = scene.getObjectByName(this.frames[this.currentFrame][17][1]).matrixWorld.elements[14];
					if(this.frames[this.currentFrame][17][0] == 1){
						this.shape.getObjectByName(linkName).add(scene.getObjectByName(this.frames[this.currentFrame][17][1]));
						scene.getObjectByName(this.frames[this.currentFrame][17][1]).matrix.set(1, 0, 0, newFrameX - oldFrameX,
																								0, 1, 0, newFrameY - oldFrameY,
												   											    0, 0, 1, newFrameZ - oldFrameZ,
																								0, 0, 0, 1);
					}else {
						scene.add(scene.getObjectByName(this.frames[this.currentFrame][17][1]));
						scene.getObjectByName(this.frames[this.currentFrame][17][1]).matrix.set(1, 0, 0, oldFrameX,
																								0, 1, 0, oldFrameY,
																								0, 0, 1, oldFrameZ,
																								0, 0, 0, 1);
					}
				}
				this.currentFrame = this.currentFrame + 1;
			}		
		}
	}
	//Function that creates 6DOF
	function sDOF(robotFrames){
		let linkInfo6DOF = [[ 1.570, -1.570,  0.000,  0.000,  0,  0.000], // "theta" rotation in z
					        [ 0.335,  0.000,  0.000, -0.405,  0.000, -0.080], // "d" translation in z
					        [-1.570,  0.000,  1.570, -1.570,  1.570,  3.141], // "alfa" rotation in x
					        [ 0.075,  0.365,  0.090,  0.000,  0.000,  0.000], // "a" translation in x
					        [ 0,  0,  0,  0,  0,  2,  0]];// kind of link
		sDOF = new Robot(linkInfo6DOF, robotFrames);
		const objLoader = new OBJLoader();
		objLoader.load('https://raw.githubusercontent.com/SetpointCapybara/kukakr5/main/models/Base.obj', (root) => {
            root.scale.set(0.001,0.001,0.001);
			root.rotation.x = 3.14/2;
			sDOF.shape.getObjectByName("base").getObjectByProperty("type", "AxesHelper").visible = false;
			sDOF.shape.getObjectByName("base").getObjectByProperty("type", "Mesh").visible = false;
            sDOF.shape.getObjectByName("base").add(root);
        });
		objLoader.load('https://raw.githubusercontent.com/SetpointCapybara/kukakr5/main/models/Axis1.obj', (root) => {
            root.scale.set(0.001,0.001,0.001);
			root.rotation.x = 3.14/2;
			root.position.set(0, 0, 0.203);
			root.rotation.y = 3.14/2;
			sDOF.shape.getObjectByName("link1").getObjectByProperty("type", "AxesHelper").visible = false;
			sDOF.shape.getObjectByName("link1").getObjectByProperty("type", "Mesh").visible = false;
            sDOF.shape.getObjectByName("link0").add(root);
        });
		objLoader.load('https://raw.githubusercontent.com/SetpointCapybara/kukakr5/main/models/Axis2.obj', (root) => {
            root.scale.set(0.001,0.001,0.001);
			root.rotation.x = -3.14/2;
			root.rotation.z = 3.14;
			root.rotation.y = -3.14/13;
			root.position.set(0, 0, 0.1);
			sDOF.shape.getObjectByName("link2").getObjectByProperty("type", "AxesHelper").visible = false;
			sDOF.shape.getObjectByName("link2").getObjectByProperty("type", "Mesh").visible = false;
            sDOF.shape.getObjectByName("link1").add(root);
        });
		objLoader.load('https://raw.githubusercontent.com/SetpointCapybara/kukakr5/main/models/Axis3.obj', (root) => {
            root.scale.set(0.001,0.001,0.001);
			root.rotation.x = 3.14/2;
			root.rotation.z = -3.14/2;
			root.position.set(0, 0, 0);
			sDOF.shape.getObjectByName("link3").getObjectByProperty("type", "AxesHelper").visible = false;
			sDOF.shape.getObjectByName("link3").getObjectByProperty("type", "Mesh").visible = false;
            sDOF.shape.getObjectByName("link2").add(root);
        });
		objLoader.load('https://raw.githubusercontent.com/SetpointCapybara/kukakr5/main/models/Axis4.obj', (root) => {
            root.scale.set(0.001,0.001,0.001);
			root.rotation.z = -3.14/2;
			root.rotation.x = 3.14/2;
			root.position.set(0.0, 0.0, -0.218);
			sDOF.shape.getObjectByName("link4").getObjectByProperty("type", "AxesHelper").visible = false;
			sDOF.shape.getObjectByName("link4").getObjectByProperty("type", "Mesh").visible = false;
            sDOF.shape.getObjectByName("link3").add(root);
        });
		objLoader.load('https://raw.githubusercontent.com/SetpointCapybara/kukakr5/main/models/Axis5.obj', (root) => {
            root.scale.set(0.001,0.001,0.001);
			root.rotation.x = 3.14/2;
			root.position.set(0.0, 0.0, 0.0);
			sDOF.shape.getObjectByName("link5").getObjectByProperty("type", "AxesHelper").visible = false;
			sDOF.shape.getObjectByName("link5").getObjectByProperty("type", "Mesh").visible = false;
            sDOF.shape.getObjectByName("link4").add(root);
        });
		objLoader.load('https://raw.githubusercontent.com/SetpointCapybara/kukakr5/main/models/Axis6o.obj', (root) => {
            root.scale.set(0.001,0.001,0.001);
			root.rotation.x = 3.14/2;
			root.position.set(0.00, 0.0, -0.012);
			//sDOF.shape.getObjectByName("link6").getObjectByProperty("type", "AxesHelper").visible = false;
			sDOF.shape.getObjectByName("link6").getObjectByProperty("type", "Mesh").visible = false;
            sDOF.shape.getObjectByName("link6").add(root);
        });
		return sDOF
	}
	//------------------------------------------------------------
	
	//--------------- BASIC ELEMENTS OF ANY SCENE ----------------
	Object3D.DefaultUp = new Vector3(0,0,1); //Pointing Z axis up
	const canvas = document.querySelector('#scene');// Selecting canvas
	const scene = new Scene();//Instantiate the Scene
	scene.background = new Color('Gainsboro');//Set background color
	const camera = new PerspectiveCamera(35, canvas.clientWidth/canvas.clientHeight, 0.1, 100);//Instantiate a camera
	camera.position.set(4, 4, 3);//Put camera in its place
	const ambientLight = new HemisphereLight('white','darkslategrey', 3,);//Instantiate Ambient light
	const controls = new OrbitControls(camera, canvas);	//Instantiate orbit controls
	controls.target.set(0, 0, 0);//Point camera at the origin
	const renderer = new WebGLRenderer({canvas, antialias: true});//Instantiate renderer
	renderer.physicallyCorrectLights = true;//Enable physically Correct Lights
	renderer.setSize(canvas.clientWidth, canvas.clientHeight);//Set render size
	renderer.setPixelRatio(window.devicePixelRatio);//Set pixel ratio
	function fitWindow() { //Function that makes scene fit browser window
		renderer.setSize(window.innerWidth, window.innerHeight);
		camera.aspect = window.innerWidth / window.innerHeight;
		camera.updateProjectionMatrix();
	}
	let sceneElements = [];
	const axesHelper = new AxesHelper( 0.5 ); //Create axis helper
	axesHelper.renderOrder = 1;
	const gridHelper = new GridHelper( 3, 6);//Create grid helper
	gridHelper.rotation.x = 3.14/2;
	scene.add(ambientLight);
	scene.add( axesHelper );
	scene.add( gridHelper );
	//------------------------------------------------------------
	
	//--------------- ADDING ELEMENTS TO THIS SCENE ---------------
	//USER INPUT GOES HERE
	
	// add stuff to the scene
	for(let i = 0; i < sceneElements.length; i++){
		scene.add(sceneElements[i].shape);
	}
	//------------------------------------------------------------
	
	
	//-------------------- THE ANIMATION LOOP -------------------
	renderer.setAnimationLoop(() => {
		//fitWindow()
		
		//MAGIC HAPPENS HERE!!!
		for(let i = 0; i < sceneElements.length; i++){
			sceneElements[i].nextFrame();
		}

		renderer.render(scene, camera);
	});
	//------------------------------------------------------------
</script>
'''
  #######################################
  # Attributes
  #######################################
  


  @property
  def listOfObjects(self):
    """A list of all objects."""
    return self._listOfObjects

  @property
  def listOfNames(self):
    """A list of all object names."""    
    return self._listOfNames

  #######################################
  # Constructor
  #######################################
    
  def __init__(self):
    self._listOfObjects=[]
    self._listOfNames=[]
    self._code = Simulation._STRJAVASCRIPT

  #######################################
  # Std. Print
  #######################################

  def __repr__(self):

    string = "Simulation: \n\n"
    string += " Variables: \n"
    string += str(self.listOfVariables)

    return string


  #######################################
  # Methods
  #######################################
    
  def run(self):
    """Run simulation."""
    self._code = Simulation._STRJAVASCRIPT
    for obj in self.listOfObjects:
      obj.genCode()
      self._code = re.sub("//USER INPUT GOES HERE", obj.code, self.code)

    display(HTML(self.code))
 
  def add(self, objSim):
    """
    Add an object to the simulation.

    Parameters
    ----------
    obj : SimObject
        The object to be added.
    """  

    if objSim.name in self.listOfNames:
      raise Exception("The name '"+objSim.name+"' is already in the list of symbols!")

    self._listOfNames.append(objSim.name)
    self._listOfObjects.append(objSim)









from utils import *
import numpy as np

def _FK(self, q = None, axis = 'eef', htm0 = None):

  if q is None:
    q = self.q
  if htm0 is None:
    htm0 = self.htm


  #GAMBIARRA
  q[5]=-q[5]


  n = len(self._links)

  #Error handling
  if not Utils.isaVector(q,n):
    raise Exception("The optional parameter 'q' should be a "+str(n)+" dimensional vector")

  if not (axis=="eef" or axis=="dh" or axis=="com"):
    raise Exception("The optional parameter 'axis' should be one of the following strings:\n"\
    "'eef': End-effector \n"\
    "'dh': All "+str(n)+" axis of Denavit-Hartenberg\n"\
    "'com': All "+str(n)+" axis centered at the center of mass of the objects")

  if not Utils.isaMatrix(htm0,4,4):
    raise Exception("The optional parameter 'htm0' should be a 4x4 homogeneous transformation matrix")
  #end error handling

  q = q + self._dq0

  T = np.zeros((n,4,4))

  for i in range(n):
    if i == 0:
      T[i,:,:] = htm0
    else:
      T[i,:,:] = T[i-1,:,:]

    if self._links[i].jointType==0:
        T[i,:,:] = T[i,:,:] @ Utils.rotz(q[i])
    else:
        T[i,:,:] = T[i,:,:] @ Utils.rotz(self._links[i].theta)

    if self._links[i].jointType==1:
        T[i,:,:] = T[i,:,:] @ Utils.trn([0,0,q[i]])
    else:
        T[i,:,:] = T[i,:,:] @ Utils.trn([0,0,self._links[i].d])        

    T[i,:,:] = T[i,:,:] @ Utils.rotx(self._links[i].alpha)
    T[i,:,:] = T[i,:,:] @ Utils.trn([self._links[i].a,0,0])

  if axis=='com':
    for i in range(n):
      T[i,3,0:3] = T[i,3,0:3] + T[i,0:3,0:3] @ self._links[i].centerShift      

  if axis=='eef':
    T = T[-1,:,:]

  return T 


from utils import *
import numpy as np
from ._dynM import _dynM
from ._dynC import _dynC
from ._dynG import _dynG

#Dynamic parameter All
def _dynParam(self, qdot, q = None):

  if q is None:
    q = self.q

  #Error handling
  if not Utils.isaVector(qdot,n):
    raise Exception("The parameter 'qdot' should be a "+str(n)+" dimensional vector")

  if not Utils.isaVector(q,n):
    raise Exception("The optional parameter 'q' should be a "+str(n)+" dimensional vector")
  #end error handling  

  Jc, FKc = self.jacGeo(q,'com')

  M = self._dynM(q, Jc, FKc)
  C = self._dynC(qdot, q, Jc, FK)
  G = self._dynG(q, Jc, FK)

  return M, C, G 


import numpy as np



class _DistStruct:

  #######################################
  # Attributes
  #######################################
  
	@property
	def idObj(self):
		"""Return the memory address for the associated object."""
		return self._idObj

	@property
	def idRobot(self):
		"""Return the memory address for the associated robot."""
		return self._idRobot

	@property
	def jacDistMat(self):
		"""
		Return the matrix in which each row we have the distance Jacobian (gradient) for each robot link.
		"""    
		return self._jacDistMat

	@property
	def distVect(self):
		"""
		Return the vector in which each row we have the distance for each robot link.
		"""       
		return np.array(self._distVect).reshape((self.noItems,))

	@property
	def noItems(self):
		"""Return the number of items."""
		return self._noItems


	def __getitem__(self, key):
		return self._listDict[key]

  #######################################
  # Constructor
  #######################################
      
	def __init__(self, obj, robot):

		self._idObj = id(obj)
		self._idRobot = id(robot)
		self._objName = obj.name
		self._robotName = robot.name
		self._noItems = 0

		n = len(robot.links)  		
		self._listDict = [];
		self._jacDistMat = np.zeros((0,n))
		self._distVect = []


  #######################################
  # Std. Print
  #######################################

	def __repr__(self):
    
		return "Distance struct between robot '"+self._robotName+"' and object '"\
    +self._objName+"', with "+str(self.noItems)+" items"



  #######################################
  # Methods
  #######################################
  
	def _append(self, i, j, d, pObj, pObjCol, Jd):
		self._listDict.append({
        "linkNumber": i,
        "linkColObjNumber": j,
        "hgDistance": d,
        "pObj": pObj,
        "pObjCol": pObjCol,
        "jacDist": Jd
    })

		self._jacDistMat = np.vstack((self._jacDistMat, Jd))
		self._distVect.append(d)
		self._noItems+=1

	def getItem(self, i, j):
		for d in self._listDict:
			if i==d["linkNumber"] and j==d["linkColObjNumber"]:		
				return d

		raise Exception("Item not found!")  

	def getClosestItem(self):
		dmin=float('inf')
		imin=-1
		for i in range(self._noItems):
			if self[i]["hgDistance"]<dmin:
				dmin=self[i]["hgDistance"]
				imin=i
    
		return self[imin]




def _genCode(self):

  self._strFrames = str(self._frames)

  #Inject code

  if self._rtype == 'generic':
    self.code = '''
    const ''' + self._strName + ''' = new Robot(''' + self._strlinkInfo + ''',''' + self._strFrames + ''');
    sceneElements.push(''' + self._strName  + ''');
    //USER INPUT GOES HERE
    '''

  if self._rtype == 'sDOF':
    self.code = '''
    const ''' + self._strName + ''' = sDOF(''' + self._strFrames + ''');
    sceneElements.push(''' + self._strName + ''');
    //USER INPUT GOES HERE
    '''


from math import *
import numpy as np
from colour import Color
import plotly.express as px

class Utils:
    """A library that contains some utilities for jupyterbox"""

  #######################################
  # Constants
  #######################################

    _PI  = 3.1415926
    _SQRTHALFPI = 1.2533141
    _SQRT2 = 1.4142135
    _CONSTJA = 2.7889
    _CONSTI0HAT1 = 0.24273
    _CONSTI0HAT2 = 0.43023

  #######################################
  # Basic functions
  #######################################

    def S(v):
      """
      Returns a 3x3 matrix that implements the cross product for a 3D vector  
      as a matricial product, that is, a matrix S(v) such that for any other 
      3D column  vector w, S(v)w = cross(v,w).
      
      Parameters
      ----------
      v : a 3D vector
          The vector for which the S matrix will be created.

      Returns
      -------
      S : 3x3 numpy matrix
          A matrix that implements the cross product with v.
      """       
      vv = np.reshape(v,(3,))
      return np.array([[ 0   , -vv[2] ,  vv[1]],
                       [ vv[2],  0    , -vv[0]],  
                       [-vv[1],  vv[0] ,  0   ]])


    def rot(axis, angle):
      """
      Homogeneous transformation matrix that represents the rotation of an
      angle in an axis.
      
      Parameters
      ----------
      axis : a 3D vector
          The axis of rotation.
      
      angle: float
          The angle of rotation, in radians.

      Returns
      -------
      T : 4x4 numpy matrix
          The homogeneous transformation matrix.
      """          
      a = np.reshape( axis,(3,))
      a = a/np.linalg.norm(a)
      K = Utils.S(a)
      Q = np.identity(3)+sin(angle)*K + (1-cos(angle))*(K @ K)
      return np.hstack([np.vstack([Q, np.array([0,0,0])]),np.array([[0],[0],[0],[1]])])

    def trn(vector):
      """
      Homogeneous transformation matrix that represents the displacement
      of a vector
      
      Parameters
      ----------
      vector : a 3D vector
          The displacement vector.
      
      Returns
      -------
      T : 4x4 numpy matrix
          The homogeneous transformation matrix.
      """      
      v = np.reshape(vector,(3,))
      return np.hstack([np.vstack([np.identity(3), 
            np.array([0,0,0])]),np.array([[v[0]],[v[1]],[v[2]],[1]])])

    def rotx(angle):
      """
      Homogeneous transformation matrix that represents the rotation of an
      angle in the 'x' axis.
      
      Parameters
      ----------
      angle: float
          The angle of rotation, in radians.

      Returns
      -------
      T : 4x4 numpy matrix
          The homogeneous transformation matrix.
      """      
      return Utils.rot([ 1,0,0] , angle )
    
    def roty(angle):
      """
      Homogeneous transformation matrix that represents the rotation of an
      angle in the 'y' axis.
      
      Parameters
      ----------
      angle: float
          The angle of rotation, in radians.

      Returns
      -------
      T : 4x4 numpy matrix
          The homogeneous transformation matrix.
      """        
      return Utils.rot( [0,1,0] , angle )   
  
    def rotz(angle):
      """
      Homogeneous transformation matrix that represents the rotation of an
      angle in the 'z' axis.
      
      Parameters
      ----------
      angle: float
          The angle of rotation, in radians.

      Returns
      -------
      T : 4x4 numpy matrix
          The homogeneous transformation matrix.
      """        
      return Utils.rot( [0,0,1] , angle)


    def axisAngle(T):
      """
      Given an homogeneous transformation matrix representing a rotation, 
      return the rotation axis angle.
      
      Parameters
      ----------
      T: 4X4 numpy array or nested list 
          Homogeneous transformation matrix of the rotation.

      Returns
      -------
      axis : 3D numpy vector
          The rotation axis.

      angle : float
          The rotation angle, in radians.        
      """        
      Q = T[0:3,0:3]
      trace = Q[0,0]+Q[1,1]+Q[2,2]
      angle = acos((trace-1)/2)
      G = Q @ Q -2*cos(angle)*Q + np.identity(3)
      ok = False
      while not ok:
          v = np.random.uniform(-100,100, size=(3,))
          w = np.random.uniform(-100,100, size=(3,))
          r = G @ v
          nr = np.linalg.norm(r)
          prod = np.transpose(w) @ r
          if nr>0.01:
            ortr = w - prod * r/(nr*nr)
            axis = Utils.S(ortr) @ ( Q @ ortr)
            naxis = np.linalg.norm(axis)
            ok = naxis > 0.01

      axis = axis/naxis
      return axis, angle

    def dpinv(A,eps):
      """
      Compute the damped pseudoinverse of the matrix A.
      
      Parameters
      ----------
      A: nxm numpy array
          The matrix to compute the damped pseudoinverse.
      
      eps: positive float
          The damping factor.
      Returns
      -------
      pinvA: mxn numpy array
          The damped pseudoinverse of A.    
      """     
      n = len(A[0,:])
      return np.linalg.inv( np.transpose(A) @ A + eps*np.identity(n)) @ np.transpose(A)

  #######################################
  # Type check functions
  #######################################

    def isaNumber(x):
      """
      Check if the argument is a float or int number
      
      Parameters
      ----------
      x: object
          Object to be verified.
      
      Returns
      -------
      isType: boolean
          If the object is of the type.   
      """   

      return str(type(x)) == "<class 'int'>" or str(type(x)) == "<class 'float'>"

    def isaNaturalNumber(x):
      """
      Check if the argument is a natural number (integer and >=0)
      
      Parameters
      ----------
      x: object
          Object to be verified.
      
      Returns
      -------
      isType: boolean
          If the object is of the type.   
      """   

      return str(type(x)) == "<class 'int'>" and x >= 0

    def isaMatrix(x , n = None, m = None):
      """
      Check if the argument is a nxm matrix of floats.
      
      Parameters
      ----------
      x: object
          Object to be verified.

      n: positive int
          Number of rows
          (default: it does not matter).

      m: positive int
          Number of columns
          (default: it does not matter).

      Returns
      -------
      isType: boolean
          If the object is of the type.   
      """  

      value = str(type(x)) == "<class 'numpy.ndarray'>" or str(type(x)) == "<class 'list'>"

      if str(type(x)) == "<class 'numpy.ndarray'>":
        shp = np.shape(x)
        if len(shp)==1:
          if (not n is None) and (not m is None): 
            value = (shp[0]==n and m==1) or (shp[0]==m and n==1)

          if (not n is None) and (m is None): 
            value = (shp[0]==n) or (n==1)

          if (n is None) and (not m is None): 
            value = (m==1) or (shp[0]==m)

          if (n is None) and (m is None): 
            value = True

        if len(shp)==2:
          if (not n is None) and (not m is None): 
            value = shp[0]==n and shp[1]==m

          if (not n is None) and (m is None): 
            value = shp[0]==n 

          if (n is None) and (not m is None): 
            value = shp[1]==m

          if (n is None) and (m is None): 
            value = True

        if len(shp)>2:
          value = False

      if str(type(x)) == "<class 'list'>":
        return Utils.isaMatrix(np.array(x),n,m)
      else:
        return value

    def isaVector(x , n= None):
      """
      Check if the argument is a n vector of floats.
      
      Parameters
      ----------
      x: object
          Object to be verified.

      n: positive int
          Number of elements
          (default: it does not matter).

      Returns
      -------
      isType: boolean
          If the object is of the type.   
      """
      return Utils.isaMatrix(x,n,1) or Utils.isaMatrix(x,1,n)

    def isaPDMatrix(x , n = None):
      """
      Check if the argument is a symmetric nxn positive (semi)-definite matrix.
      
      Parameters
      ----------
      x: object
          Object to be verified.

      n: positive int
          Dimension of the square matrix
          (default: it does not matter).
    
      Returns
      -------
      isType: boolean
          If the object is of the type.   
      """
      value = Utils.isaMatrix(x,n,n)

      if value:
        value = np.allclose(x, np.transpose(x), rtol=1e-05, atol=1e-08)

      if value:
        try:
          np.linalg.cholesky(x)
        except:
          value=False
      
      return value

    def isaColor(color):
      """
      Check if the argument is a HTML-compatible string that represents a color.
      
      Parameters
      ----------
      x: object
          Object to be verified.
      
      Returns
      -------
      isType: boolean
          If the object is of the type.   
      """

      try:
        color = color.replace(" ", "")
        Color(color)
        return True
      except ValueError: 
        return False

    def isaSimpleObject(obj):
      """
      Check if the argument is a ball, box or cylinder.
      
      Parameters
      ----------
      x: object
          Object to be verified.
      
      Returns
      -------
      isType: boolean
          If the object is of the type.   
      """      
      try:
        return obj._isaSimpleObject
      except:
        return False

  #######################################
  # Plotting functions
  #######################################

    def plot(xv , yv, title="", xname="x", yname="y", labels="" ):

      fig = px.line(width=800, height=400)

      #Error handling


      if not Utils.isaVector(xv):
        raise Exception("The parameter 'xv' should be a vector")

      m = max(np.shape(xv))

      if not Utils.isaMatrix(yv,None,m):
        raise Exception("The parameter 'yv' should be a matrix with "+str(m)+" columns.")

      n = 1 if len(np.shape(yv))==1 else np.shape(yv)[0]

      listNames=[]

      if str(type(labels)) == "<class 'str'>":
        for i in range(n):
          listNames.append(labels+"_"+str(i+1))

      else:
        if str(type(labels)) == "<class 'list'>" and len(labels)==n:
          for i in range(n):
            if str(type(labels[i])) == "<class 'str'>":
              listNames.append(labels[i])
            else:
              raise Exception("Optional parameter 'labels' must be either a string or a list of "+str(n)+" strings")
        else:
          raise Exception("Optional parameter 'labels' must be either a string or a list of "+str(n)+" strings")


      #end error handling
      if n>1:
        for i in range(n):
          fig.add_scatter(x=xv, y=yv[i], mode="lines", name=listNames[i])
      else:
          fig.add_scatter(x=xv, y=yv, mode="lines", name=listNames[0])

      fig.update_xaxes(title_text=xname)
      fig.update_yaxes(title_text=yname)
      fig.show()

      return fig

  #######################################
  # Distance computation functions
  #######################################

    def _funJ(u):
      return Utils._CONSTJA/((Utils._CONSTJA-1)*sqrt(Utils._PI*u*u) + sqrt(Utils._PI*u*u+Utils._CONSTJA*Utils._CONSTJA))

    def funInt( v, h, L):
      v = abs(v)
      if v<=L:
        A1 = exp(-(L-v)*(L-v)/(2*h*h))*Utils._funJ((v-L)/(Utils._SQRT2*h))
        A2 = exp(-(L+v)*(L+v)/(2*h*h))*Utils._funJ((v+L)/(Utils._SQRT2*h))
        return -h*h*log(Utils._SQRTHALFPI*(h/(2*L))*(2-A1-A2))
      else:
        A1 = Utils._funJ((v-L)/(Utils._SQRT2*h))
        A2 = exp(-2*L*v/(h*h))*Utils._funJ((v+L)/(Utils._SQRT2*h))
        return 0.5*(v-L)*(v-L) -h*h*log(Utils._SQRTHALFPI*(h/(2*L))*(A1-A2))

    def _funI0hat(u):
      return pow(1+0.25*u*u,-0.25)*(1 + Utils._CONSTI0HAT1*u*u)/(1 + Utils._CONSTI0HAT2*u*u)

    def _funf(nu, rho):
      A1 = exp(-0.5*(rho-nu)*(rho-nu))
      A2 = exp(-0.5*(rho+nu)*(rho+nu))
      return rho*(A1+A2)*Utils._funI0hat(rho*nu)

    def _funfhat(nu, rho, rhobar):
      A1 = exp(-0.5*(rho-nu)*(rho-nu) + 0.5*(rhobar-nu)*(rhobar-nu))
      A2 = exp(-0.5*(rho+nu)*(rho+nu) + 0.5*(rhobar-nu)*(rhobar-nu))
      return rho*(A1+A2)*Utils._funI0hat(rho*nu)

    def funCir(v, h, R):

      v = abs(v) 
      N=7
      node= [0.94910, -0.74153, -0.40584, 0, 0.40584, 0.74153, 0.94910]
      weight= [0.12948, 0.27970, 0.38183, 0.4179, 0.38183, 0.27970, 0.12948]

      if v <= R:
        F_low = max(0,sqrt((v/h)*(v/h)+1)-3)
        F_up = min(R/h, sqrt((v/h)*(v/h)+1)+3)
        delta = 0.5*(F_up-F_low)
        y=0
        for i in range(N):
          y = y + weight[i]*Utils._funf(v/h,F_low + delta*(node[i]+1))

        y = delta * y
        return -h*h*log(y*(h/R)*(h/R))
      else:
        F_low = 0
        F_up = R/h
        delta = 0.5*(F_up-F_low)
        rhobar = F_low + delta *(node[N-1]+1)
        y=0;
        for i in range(N):
          y = y + weight[i]*Utils._funfhat(v/h,F_low + delta *(node[i]+1) , rhobar)

        y = delta * y
        return 0.5*(v-h*rhobar)*(v-h*rhobar)-h*h*log(y*(h/R)*(h/R))
      
    def funSph( v, h, R):  

      v = abs(v);
      C = 3*(h*h)/(2*R*R*R)

      if v <= R :
        if v==0:
          return -h*h*log(C*(-2*R*exp(-(R*R)/(2*h*h)) + 2*R*exp(-Utils.funInt(0,h,R)/(h*h))))
        else:
          A1 = exp(-((R+v)*(R+v)/(2*h*h)))
          A2 = exp(-((R-v)*(R-v)/(2*h*h)))
          return -h*h*log(C*(h*h*(A1-A2)/v + 2*R*exp(-Utils.funInt(v,h,R)/(h*h))))
   
      else:
        A1 = exp(-(2*R*v/(h*h)))
        A2 = 1
        return 0.5*(v-R)*(v-R)-h*h*log(C*(h*h*(A1-A2)/v + 2*R*exp((0.5*(v-R)*(v-R)-Utils.funInt(v,h,R))/(h*h))))  

    def computehgDist(objA, objB, h=0.001, g = 0.001, pA = None, tol = 0.001, maxIter=20):
    
      if pA is None:
        pA = np.random.uniform(-3,3, size=(3,))

      #Error handling

      if not Utils.isaSimpleObject(objA):
        raise Exception("The parameter 'objA' must be either a box, a ball or a cylinder")

      if not Utils.isaSimpleObject(objB):
        raise Exception("The parameter 'objB' must be either a box, a ball or a cylinder")

      if not Utils.isaNumber(h) or h<=0:
        raise Exception("The optional parameter 'h' must be nonnegative number")

      if not Utils.isaNumber(g) or g<=0:
        raise Exception("The optional parameter 'g' must be nonnegative number")

      if not Utils.isaVector(pA,3):
        raise Exception("The optional parameter 'pA' must be a 3D vector")

      if not Utils.isaNumber(tol) or tol<=0:
        raise Exception("The optional parameter 'tol' must be a nonnegative number")

      if not Utils.isaNaturalNumber(maxIter):
        raise Exception("The optional parameter 'maxIter' must be a nonnegative integer")

      #end error handling

      algEnd=False
      i=0

      while (not algEnd) and i < maxIter:
        pAant = pA
        pB, dpA = objB.hProjection(pA,g)
        pA, dpB = objA.hProjection(pB,h)
        algEnd = np.linalg.norm(pA-pAant) <tol
        i+=1

      dist = np.linalg.norm(pA-pB)
      d = sqrt(2*(dpA+dpB - 0.5*dist*dist))

      return pA, pB, d
      


from utils import *
import numpy as np

#(Private function) Used in IK
def _evolveConfig(self, q, ptol, atol, FK, iterRemain):

  n = len(self.links)
  found = False
  zerou = False
  iterEnd = False

  dt=min(ptol,0.005*atol,0.01)
  K=2
  eps=0.001
  tolu=0.0001/dt
  i=0

  while (not found) and (not zerou) and (not iterEnd):

    r, Jr = self.taskFunction(FK, np.array(q))

    #pD = FK[0:3,3]
    #xD = FK[0:3,0]
    #yD = FK[0:3,1]
    #zD = FK[0:3,2]

    #J,FKC = self.jacGeo(np.array(q))
    #p = FKC[0:3,3]
    #x = FKC[0:3,0]
    #y = FKC[0:3,1]
    #z = FKC[0:3,2]  

    #r = np.zeros((n,))
    #r[0:3] = p-pD
    #r[3] = sqrt(max(1 - np.transpose(xD) @ x,0))
    #r[4] = sqrt(max(1 - np.transpose(yD) @ y,0))
    #r[5] = sqrt(max(1 - np.transpose(zD) @ z,0))

    #Jr = np.zeros((6,n))
    #Jr[0:3,:] = J[0:3,:]
    #Jr[3,:] = np.transpose(xD) @ Utils.S(x) @ J[3:6,:]
    #Jr[4,:] = np.transpose(yD) @ Utils.S(y) @ J[3:6,:]
    #Jr[5,:] = np.transpose(zD) @ Utils.S(z) @ J[3:6,:]

    u = Utils.dpinv(Jr,eps) @ (-K*r)
    q = q + u*dt

    epos = max(abs(r[0:3]))
    eori = max([57.29577*acos(min(max(1-num*num,-1),1)) for num in r[3:6]])
    i+=1

    found = (epos < ptol) and (eori < atol)
    zerou = max(abs(u))<tolu
    iterEnd = i > iterRemain


  return found, i, q

#Inverse kinematics for the end-effector
def _IK(self, htm, q0 = None, ptol=0.005, atol=5, noIterMax=2000):

  n = len(self._links)
  if q0 is None:
    q0 = 6.2831*np.random.rand(n)-3.1416  

  #Error handling
  if not Utils.isaMatrix(htm,4,4):
    raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix")

  if not Utils.isaNaturalNumber(noIterMax):
    raise Exception("The optional parameter 'noIterMax' should be a nonnegative integer number")  

  if not Utils.isaVector(q0,n):
    raise Exception("The optional parameter 'q0' should be a "+str(n)+" dimensional vector")

  if (not Utils.isaNumber(ptol)) or ptol<=0:
    raise Exception("The optional parameter 'pol' should be a nonnegative number")

  if (not Utils.isaNumber(atol)) or atol<=0:
    raise Exception("The optional parameter 'atol' should be a nonnegative number")

  
  #end error handling   

  j=0
  found = False
  q=q0
  noIterRemain = noIterMax

  while not found and noIterRemain>=0:
    found, i, q = _evolveConfig(self,q,ptol,atol,htm,noIterRemain)
    noIterRemain -= i
    if not found:
      q = 6.2831*np.random.rand(n)  

  if not found:     
    raise Exception("Solution for IK not found. You can try the following: \n"\
    " Increasing the maximum number of iterations, 'noIterMax' (currently "+str(noIterMax)+")\n"\
    " Increasing the tolerance for the position, 'ptol' (currently "+str(ptol)+" meters)\n"\
    " Increasing the tolerance for the orientation, 'atol' (currently "+str(atol)+" degrees)")
  else:
    return q



from utils import *
import numpy as np

#Function used for task function/task Jacobian
def _taskFunction(self, htmd, q = None, htm = None):

  if q is None:
    q = self.q

  if htm is None:
    htm = self.htm

  #Error handling
  if not Utils.isaMatrix(htmd,4,4):
    raise Exception("The parameter 'htmd' should be a 4x4 homogeneous transformation matrix")
  #end error handling  

  pdes = htmd[0:3,3]
  xdes = htmd[0:3,0]
  ydes = htmd[0:3,1]
  zdes = htmd[0:3,2]

  J, FK = self.jacGeo(q, MTH0 = htm)
  p = FK[0:3,3]
  x = FK[0:3,0]
  y = FK[0:3,1]
  z = FK[0:3,2]

  r = np.zeros((6,))
  r[0:3] = p-pdes
  r[3] = sqrt(max(1 - np.transpose(xdes) @ x,0))
  r[4] = sqrt(max(1 - np.transpose(ydes) @ y,0))
  r[5] = sqrt(max(1 - np.transpose(zdes) @ z,0))

  n = len(self.links)
  Jr = np.zeros((6,n))
  Jr[0:3,:] = J[0:3,:]
  Jr[3,:] = np.transpose(xdes) @ Utils.S(x) @ J[3:6,:]
  Jr[4,:] = np.transpose(ydes) @ Utils.S(y) @ J[3:6,:]
  Jr[5,:] = np.transpose(zdes) @ Utils.S(z) @ J[3:6,:]

  return r, Jr



from utils import *
import numpy as np

#Dynamic parameter Inertia Matrix
def _dynM(self, q = None):

  if q is None:
    q = self.q
  
  n = len(self._links)

  #Error handling
  if not Utils.isaVector(q,n):
    raise Exception("The optional parameter, 'q', should be a "+str(n)+" dimensional vector")
  #end error handling

  Jc, FKc = self.jacGeo(q,'com')

  return self._dynMaux(q,Jc,FKc)

def _dynMaux(self, q, Jc, FKc):

  n = len(self._links)
  M = np.zeros((n,n))

  for i in range(n):
    M = M + self._links[i].mass * np.transpose(Jc[i,0:3,:]) @ Jc[i,0:3,:]
    M = M + np.transpose(Jc[i,3:6,:]) @ FKc[i,0:3,0:3] @ \
            self._links[i].inertiaMatrix @ \
            np.transpose(FKc[i,0:3,0:3]) @ Jc[i,3:6,:]

  return M



from utils import *
import numpy as np

#Add config to animation queue
def _addConfig(self, q, htm = None):


  if htm is None:
    htm = self.htm

  n = len(self._linkInfo[0])

  #Error handling
  if not Utils.isaVector(q,n):
    raise Exception("The parameter 'q' should be a "+str(n)+" dimensional vector")

  if not Utils.isaMatrix(htm,4,4):
    raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix")
  #end error handling

  self._q = np.array(q)

  f = [htm[0][0], htm[0][1], htm[0][2], htm[0][3],
        htm[1][0], htm[1][1], htm[1][2], htm[1][3],
        htm[2][0], htm[2][1], htm[2][2], htm[2][3],
        0        ,         0,         0,         1, np.ndarray.tolist(q) ]


  self._htm = htm
  self._frames.append(f)
  



from utils import *
import numpy as np

#Set config. Restart animation queue
def _setConfig(self, q = None, htm = None):

  if q is None:
    q= self.q0

  if htm is None:
    htm = self.htm

  n = len(self._linkInfo[0])
 
  self._frames = []
  self.code = ''
  self.addConfig(q,htm)
